#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/ctype.h>
#include <linux/string.h>
#include <asm/uaccess.h>
#include <linux/file.h>
#include <linux/smp_lock.h>

#define SHFS_DEBUG_HERE

#include "shfs.h"
#include "shfs_proc.h"
#include "shfs_proto.h"

void
shfs_init_root_dirent(struct shfs_sb_info *server, struct shfs_fattr *fattr)
{
	memset(fattr, 0, sizeof(*fattr));
	fattr->f_nlink = 1;
	fattr->f_uid = server->mnt.uid;
	fattr->f_gid = server->mnt.gid;
	fattr->f_blksize = 512;

	fattr->f_ino = 2;
	fattr->f_mtime = CURRENT_TIME;
	fattr->f_mode = server->mnt.root_mode;
	fattr->f_size = 512;
	fattr->f_blocks = 0;
}

static int
shfs_parse_mode(char *mode)
{
	return ((mode[0]-'0')<<6) + ((mode[1]-'0')<<3) + (mode[2]-'0');
}

int
shfs_parse_options(struct shfs_sb_info *info, void *opts)
{
	char *p, *q;
	int i;

	strcpy(info->mnt.root, "");
	info->mnt.pin = info->mnt.pout = NULL;
	info->mnt.fmask = 00177777;

	if (!opts) 
		return -1;

	p = strtok(opts, ",");
	for (; p; p = strtok(NULL, ",")) {
		if (strncmp(p, "root=", 5) == 0) {
			if (strlen(p+5) > SHFS_PATH_MAX)
				goto ugly_opts;
			strcpy(info->mnt.root, p+5);
		} else if (strncmp(p, "ro", 2) == 0) {
			info->mnt.readonly = 1;
		} else if (strncmp(p, "rw", 2) == 0) {
			info->mnt.readonly = 0;
		} else if (strncmp(p, "suid", 4) == 0) {
			info->mnt.fmask |= S_ISUID|S_ISGID;
		} else if (strncmp(p, "nosuid", 6) == 0) {
			info->mnt.fmask &= ~(S_ISUID|S_ISGID);
		} else if (strncmp(p, "dev", 3) == 0) {
			info->mnt.fmask |= S_IFCHR|S_IFBLK;
		} else if (strncmp(p, "nodev", 5) == 0) {
			info->mnt.fmask &= ~(S_IFCHR|S_IFBLK);
		} else if (strncmp(p, "exec", 4) == 0) {
			info->mnt.fmask |= S_IXUSR|S_IXGRP|S_IXOTH;
		} else if (strncmp(p, "noexec", 6) == 0) {
			info->mnt.fmask &= ~(S_IXUSR|S_IXGRP|S_IXOTH);
		} else if (strncmp(p, "nocache", 7) == 0) {
			info->mnt.disable_fcache = 1;
		} else if (strncmp(p, "generic", 7) == 0) {
			info->mnt.generic_host = 1;
		} else if (strncmp(p, "preserve", 8) == 0) {
			info->mnt.preserve_own = 1;
		} else if (strncmp(p, "uid=", 4) == 0) {
			if (strlen(p+4) > 5)
				goto ugly_opts;
			q = p+4;
			i = simple_strtoul(q, &q, 10);
			info->mnt.uid = i; 
		} else if (strncmp(p, "gid=", 4) == 0) {
			if (strlen(p+4) > 5)
				goto ugly_opts;
			q = p+4;
			i = simple_strtoul(q, &q, 10);
			info->mnt.gid =i; 
		} else if (strncmp(p, "rmode=", 6) == 0) {
			if (strlen(p+6) > 3)
				goto ugly_opts;
			info->mnt.root_mode = S_IFDIR | (shfs_parse_mode(p+6) & (S_IRWXU | S_IRWXG | S_IRWXO));
		} else if (strncmp(p, "fdi=", 4) == 0) {
			if (strlen(p+4) > 5)
				goto ugly_opts;
			q = p+4;
			i = simple_strtoul(q, &q, 10);
			VERBOSE("fdi=%d\n", i);
			info->mnt.pin = fget(i);
		} else if (strncmp(p, "fdo=", 4) == 0) {
			if (strlen(p+4) > 5)
				goto ugly_opts;
			q = p+4;
			i = simple_strtoul(q, &q, 10);
			VERBOSE("fdo=%d\n", i);
			info->mnt.pout = fget(i);
		}
	}
	if (!info->mnt.pin || !info->mnt.pout)
		goto ugly_opts;
	return 0;
	
ugly_opts:
	VERBOSE("Your options sucks!\n");
	return -1;
}

static int clear_garbage(struct shfs_sb_info *);

static int
pipe_write(struct shfs_sb_info *info, const void *buf, int len)
{
	int bytes, wr = 0;
	int res;
	struct file *f = info->mnt.pout;
	mm_segment_t fs;

	if (info->mnt.garbage)
		if ((res = clear_garbage(info)) < 0)
			return res;
	
	fs = get_fs();
	set_fs(get_ds());
	bytes = len;
retry:
	while (bytes && ((wr = f->f_op->write(f, buf, bytes, &f->f_pos)) > 0)) {
		DEBUG("Wrote: %d\n", wr);
		buf += wr;
		bytes -= wr;
	}
	if (wr == -EAGAIN)
		goto retry;
	set_fs(fs);
	
	if (wr < 0) {
		info->mnt.garbage = 1;
		info->mnt.garbage_write = bytes;
		return wr;
	}
	return len;
}

int 
shfs_pipe_printf(struct shfs_sb_info *info, const char *fmt, ...)
{
	va_list ap;
	int res;

	if (info->mnt.garbage)
		if ((res = clear_garbage(info)) < 0)
			return res;

	va_start(ap, fmt);
	/* shfs is locked, we can write to printf_buffer safely */
	vsnprintf(info->mnt.printf_buffer, PRINTF_BUFFER_SIZE, fmt, ap);
	va_end(ap);

	DEBUG("Write line: %s\n", info->mnt.printf_buffer);
	return pipe_write(info, (void *)info->mnt.printf_buffer, strlen(info->mnt.printf_buffer));
}

#define BUFFER info->mnt.readln_buffer
#define LEN    info->mnt.readln_buffer_len

static int
pipe_read(struct shfs_sb_info *info, void *buf, int len)
{
	int bytes, c, res, rd = 0;
	struct file *f = info->mnt.pin;
	mm_segment_t fs;

	if (info->mnt.garbage)
		if ((res = clear_garbage(info)) < 0)
			return res;

	if (LEN > 0) {
		c = len > LEN ? LEN : len;
		memcpy(buf, BUFFER, c);
		buf+= c;
		LEN -= c;
		len -= c;
		if (LEN > 0)
			memmove(BUFFER, BUFFER+c, LEN);
	}

	fs = get_fs();
	set_fs(get_ds());
	bytes = len;
	while (bytes && ((rd = f->f_op->read(f, buf, bytes, &f->f_pos)) > 0)) {
		buf += rd;
		bytes -= rd;
	}
	set_fs(fs);
	
	if (rd < 0) {
		info->mnt.garbage = 1;
		info->mnt.garbage_read = bytes;
		return rd;
	}
	return len;
}

int 
shfs_pipe_readln(struct shfs_sb_info *info, char *buf, int len)
{
	struct file *f = info->mnt.pin;
	int c, res, l = 0, rd = 0;
	char *str;
	mm_segment_t fs;

	if (info->mnt.garbage)
		if ((res = clear_garbage(info)) < 0)
			return res;

	do {
		str = memchr(BUFFER, '\n', LEN);
		if (str) {
			DEBUG("hit!\n");
			*str = '\0';
			strncpy(buf, BUFFER, len-1);
			buf[len-1] = '\0';
			c = LEN-(str-BUFFER+1);
			memmove(BUFFER, str+1, c);
			LEN = c;
		} else {
			DEBUG("miss, size: %d\n", LEN);
			c = READLN_BUFFER_SIZE - LEN;
			if (c > 0) {
				fs = get_fs();
				set_fs(get_ds());
				rd = f->f_op->read(f, BUFFER+LEN, c, &f->f_pos);
				DEBUG("rd: %d\n", rd);
				set_fs(fs);
				if (rd == -EAGAIN)
					continue;
				if (rd < 0) {
					info->mnt.garbage = 1;
					info->mnt.garbage_read = 0;
					return rd;
				}
				LEN += rd;
			} else {
				BUFFER[READLN_BUFFER_SIZE-1] = '\n';
			}
		}
		/* do not lock while reading 0 bytes */
		if (l++ > READLN_BUFFER_SIZE && LEN == 0)
			return -1;
	} while (!str);
	
	DEBUG("size: %d\n", LEN);
	DEBUG("Read line(%d): %s\n", strlen(buf), buf);
	return strlen(buf);
}

static int
reply(char *s)
{
	if (strncmp(s, "### ", 4))
		return 0;
	return simple_strtoul(s+4, NULL, 10);
}

static int
clear_garbage(struct shfs_sb_info *info)
{
	char line[SHFS_LINE_MAX];
	static unsigned long seq = 12345;
	int i, state, c, garbage;
	int res;

	DEBUG(" start\n");
	garbage = info->mnt.garbage_write;
	if (garbage)
		memset(line, ' ', SHFS_LINE_MAX);
	while (garbage > 0) {
		c = garbage < SHFS_LINE_MAX ? garbage : SHFS_LINE_MAX;
		info->mnt.garbage = 0;
		res = pipe_write(info, line, c);
		if (res < 0) {
			info->mnt.garbage_write = garbage;
			goto error;
		}
		garbage -= res;
	}
	info->mnt.garbage_write = 0;
	garbage = info->mnt.garbage_read;
	while (garbage > 0) {
		c = garbage < SHFS_LINE_MAX ? garbage : SHFS_LINE_MAX;
		info->mnt.garbage = 0;
		res = pipe_read(info, line, c);
		if (res < 0) {
			info->mnt.garbage_read = garbage;
			goto error;
		}
		garbage -= res;
	}
	info->mnt.garbage_read = 0;
		
	info->mnt.garbage = 0;
	if ((res = do_ping(info, seq++)) < 0)
		goto error;
	DEBUG("reading..\n");
	state = 0;
	for (i = 0; i < 100000; i++) {
		info->mnt.garbage = 0;
		if ((res = shfs_pipe_readln(info, line, SHFS_LINE_MAX)) < 0)
			goto error;
		if (((state == 0) || (state == 1)) && reply(line) == REP_PRELIM) {
			state = 1;
		} else if (state == 1 && simple_strtoul(line, NULL, 10) == (seq-1)) {
			state = 2;
		} else if (state == 2 && reply(line) == REP_NOP) {
			DEBUG(" cleared.\n");
			return 0;
		} else {
			state = 0;
		}
	}
error:
	info->mnt.garbage = 1;
	DEBUG(" failure!\n");
	return res;
}



unsigned int
shfs_get_this_year(void)
{
	unsigned long s_per_year = 365*24*3600;
	unsigned long delta_s = 24*3600;

	unsigned int year = CURRENT_TIME / (s_per_year + delta_s/4);
	return 1970 + year; 
}

unsigned int
shfs_get_this_month() {
	int month = -1;
	long secs = CURRENT_TIME % (365*24*3600+24*3600/4);
	static long days_month[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	if (shfs_get_this_year() % 4) {
		days_month[1] = 28;
	} else {
		days_month[1] = 29;
	}
	while (month < 11 && secs >= 0) {
		month++;
		secs-=24*3600*days_month[month];
	}
	return (unsigned int)month;
}

unsigned int
shfs_get_month(char *mon)
{
	static char* months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	unsigned int i;

	for (i = 0; i<12; i++)
		if(strcmp(mon, months[i]) == 0) {
			DEBUG("%s: %u\n", mon, i);
			return i;
		}

	return 0;
}

static int
shfs_parse_dir(char *s, char **col)
{
	int c = 0;
	int is_space = 1;
	int device = 0;

	while (*s) {
		if (c == DIR_COLS)
			break;
		
		if (*s == ' ') {
			if (!is_space) {
				if ((c-1 == DIR_SIZE) && device) {
					while (*s && (*s == ' '))
						s++;
					while (*s && (*s != ' '))
						s++;
				}
				*s = '\0';
				is_space = 1;
			}
		} else {
			if (is_space) {
				/* (b)lock/(c)haracter device hack */
				col[c++] = s;
				is_space = 0;
				if ((c-1 == DIR_PERM) && ((*s == 'b')||(*s == 'c'))) {
					device = 1;
				}

			}
		}
		s++;
	}
	return c;
}

int
shfs_loaddir(struct shfs_sb_info* info, char* name, struct shfs_directory* dir)
{
	char line[SHFS_LINE_MAX];
	char path[SHFS_PATH_MAX];
	char *col[DIR_COLS];
	char *b, *s;
	struct shfs_dir_list_node *p, *q;
	umode_t mode;
	unsigned int year, mon, day, hour, min, this_year, this_month;
	int device; 
	int namelen;
	int res;
	
	DEBUG(" reading %s\n", name);
	
	this_year = shfs_get_this_year();
	this_month = shfs_get_this_month();
	
	if (strlen(info->mnt.root) + strlen(name) + 1 > SHFS_PATH_MAX) {
		VERBOSE("Path too long!\n");
		return -1;
	}

	if ((res = do_lsdir(info, name)) < 0) {
		VERBOSE("ls request failed!\n");
		return res;
	}
	
	while ((res = shfs_pipe_readln(info, line, SHFS_LINE_MAX)) > 0) {
		switch (reply(line)) {
		case REP_COMPLETE:
			return 0;
		case REP_EPERM:
			return -EPERM;
		case REP_ENOENT:
			return -ENOENT;
		case REP_ERROR:
			return -EIO;
		}

		res = shfs_parse_dir(line, col);
		if (res != DIR_COLS)
			continue;		/* skip `total xx' line */
		
		s = col[DIR_NAME];
		if ((strcmp(s, ".") != 0) && (strcmp(s, "..") != 0)) {
			p = (struct shfs_dir_list_node*)KMEM_ALLOC("dir", dir_list_cache, GFP_KERNEL);
			if (!p) {
				VERBOSE(" slab cache alloc error!\n");
				res = -1;
				goto out;
			}
			memset(p, 0, sizeof(struct shfs_dir_list_node));

			namelen = strlen(s) + 1;
			
			if(strstr(s, "->")){
				char *ss = strstr(s, "->");
				if(*(ss + 3) == '/')
					namelen += strlen(info->mnt.mount_point);
			}
		
			if (namelen + 1 > SHFS_PATH_MAX) {
				VERBOSE("too long entry!\n");
				KMEM_FREE("dir", dir_list_cache, p);
				res = -1;
				goto out;
			}
			p->entry.name = (char *)KMEM_ALLOC("name", dir_name_cache, GFP_KERNEL);
			if (!p->entry.name) {
				VERBOSE("slab cache alloc error!\n");
				KMEM_FREE("dir", dir_list_cache, p);
				res = -1;
				goto out;
			}

			strcpy(p->entry.name, s);

			s = col[DIR_PERM];
			mode = 0; device = 0;
			switch(s[0]) {
			case 'b':
				device = 1;
				if ((info->mnt.fmask & S_IFMT) == S_IFBLK)
					mode = S_IFBLK;
				else
					mode = S_IFREG;
				break;
			case 'c':
				device = 1;
				if ((info->mnt.fmask & S_IFMT) == S_IFCHR)
					mode = S_IFCHR;
				else
					mode = S_IFREG;
				break;
			case 's':
			case 'S':			/* IRIX64 socket */
				mode = S_IFSOCK;
				break;
			case 'd':
				mode = S_IFDIR;
				break;
			case 'l':
				mode = S_IFLNK;
				break;
			case '-':
				mode = S_IFREG;
				break;
			case 'p':
				mode = S_IFIFO;
				break;
			}
			if (s[1] == 'r') mode |= S_IRUSR;
			if (s[2] == 'w') mode |= S_IWUSR;
			if (s[3] == 'x') mode |= S_IXUSR;
			if (s[3] == 's') mode |= S_IXUSR | S_ISUID;
			if (s[3] == 'S') mode |= S_ISUID;
			if (s[4] == 'r') mode |= S_IRGRP;
			if (s[5] == 'w') mode |= S_IWGRP;
			if (s[6] == 'x') mode |= S_IXGRP;
			if (s[6] == 's') mode |= S_IXGRP | S_ISGID;
			if (s[6] == 'S') mode |= S_ISGID;
			if (s[7] == 'r') mode |= S_IROTH;
			if (s[8] == 'w') mode |= S_IWOTH;
			if (s[9] == 'x') mode |= S_IXOTH;
			if (s[9] == 't') mode |= S_ISVTX | S_IXOTH;
			if (s[9] == 'T') mode |= S_ISVTX;
			p->entry.mode = (mode & S_IFMT) == S_IFREG ? mode & info->mnt.fmask : mode;

			p->entry.ino = simple_strtoul(col[DIR_INODE], NULL, 10);
//			p->entry.ino = iunique(info->sb, 2);
			p->entry.uid = simple_strtoul(col[DIR_UID], NULL, 10);
			p->entry.gid = simple_strtoul(col[DIR_GID], NULL, 10);
			
			if (!device) {
				p->entry.size = simple_strtoul(col[DIR_SIZE], NULL, 10);
			} else {
				unsigned short major, minor;
				p->entry.size = 0;
				major = (unsigned short) simple_strtoul(col[DIR_SIZE], &s, 10);
				while (*s && (!isdigit(*s)))
					s++;
				minor = (unsigned short) simple_strtoul(s, NULL, 10);
				p->entry.rdev = MKDEV(major, minor);
			}
			p->entry.nlink = simple_strtoul(col[DIR_NLINK], NULL, 10);
			p->entry.blocksize = 4096;
			p->entry.blocks = (p->entry.size + 4095) >> 11;

			mon = shfs_get_month(col[DIR_MONTH]);
			day = simple_strtoul(col[DIR_DAY], NULL, 10);
			
			s = col[DIR_YEAR];
			if (!strchr(s, ':')) {
				year = simple_strtoul(s, NULL, 10);
				hour = 12;
				min = 0;
			} else {
				year = this_year;
				if (mon > this_month) 
					year--;
				b = strchr(s, ':');
				*b = 0;
				hour = simple_strtoul(s, NULL, 10);
				min = simple_strtoul(++b, NULL, 10);
			}
			p->entry.atime = mktime(year, mon + 1, day, hour, min, 0);

			/* symlink hack */
			if (((mode & S_IFMT) == S_IFLNK) && (strstr(p->entry.name, "->"))) {
				char *ss = strstr(p->entry.name, "->");
				
				*(ss - 1) = 0;
				if ((*(ss + 3) == '/') && (strlen(info->mnt.mount_point) + strlen(ss + 3) + 1 < SHFS_PATH_MAX)) {
					strcpy(path, ss+3);
					sprintf(ss+3, "%s%s", info->mnt.mount_point, path);
				}
			}
			DEBUG("Name: %s, mode: %o, inode: %lu, size: %lu, nlink: %d, month: %d, day: %d, year: %d, hour: %d, min: %d (time: %lu)\n", p->entry.name, p->entry.mode, p->entry.ino, p->entry.size, p->entry.nlink, mon, day, year, hour, min, p->entry.atime);

			p->prev = NULL;
			p->next = dir->head;
			dir->head = p;
		}
	}
out:
	for (p = dir->head; p; p = q) {
		q = p->next;
		KMEM_FREE("name", dir_name_cache, p->entry.name);
		KMEM_FREE("dir", dir_list_cache, p);
	}
	info->mnt.garbage_read = 0;
	info->mnt.garbage = 1;
	return res;
}

/* aaa'aaa -> aaa'\''aaa */
static int
replace_quote(char *name)
{
	char *s;
	char *r = name;
	int q = 0;
	
	while (*r) {
		if (*r == '\'')
			q++;
		r++;
	}
	s = r+(3*q);
	if ((s-name+1) > SHFS_PATH_MAX)
		return -1;
	
	*(s--) = '\0';
	r--;
	while (q) {
		if (*r == '\'') {
			s -= 3;
			s[0] = '\'';
			s[1] = '\\';
			s[2] = '\'';
			s[3] = '\'';
			q--;
		} else {
			*s = *r;
		}
		s--;
		r--;
	}
	return 0;
}

int
shfs_get_name(struct dentry *d, char *name)
{
	int len = 0;
	struct dentry *p;

	for (p = d; p != p->d_parent; p = p->d_parent)
		len += p->d_name.len + 1;
		
	if (len + 1 > SHFS_PATH_MAX) 
		return -1;
	if (len == 0) {
		name[0] = '/';
		name[1] = '\0';
		return 0;
	}
	
	name[len] = 0;
	for (p = d; p != p->d_parent; p = p->d_parent) {
		len -= p->d_name.len;
		strncpy(&(name[len]), p->d_name.name, p->d_name.len);
		len--;
		name[len] = '/';
	}

	return replace_quote(name);
}

inline int
shfs_lock(struct shfs_sb_info *info)
{
	int res;
	DEBUG("Locking...\n");
	res = down_interruptible(&(info->sem));
	DEBUG("Locked.\n");
	return (res != -EINTR);
}

inline void
shfs_locku(struct shfs_sb_info *info)
{
	DEBUG("Locking...\n");
	down(&(info->sem));
	DEBUG("Locked.\n");
}

inline void
shfs_unlock(struct shfs_sb_info *info)
{
	DEBUG("Unlocking...\n");
	up(&(info->sem));
	DEBUG("Unlocked.\n");
}

inline int
shfs_remove_sigpipe(int result)
{
	switch (result) {
//	    case -ERESTARTSYS:
//	    	result = -EINTR;
//		goto cont;
	    case -EPIPE:
		result = -EIO;
//	    cont:
		if (sigismember(&current->pending.signal, SIGPIPE)) {
			sigdelset(&current->pending.signal, SIGPIPE);
			current->sigpending--;
		}
		break;
	}
	return result;
}

int
shfs_get_attr(struct dentry *dentry, struct shfs_fattr *fattr, struct shfs_sb_info *info){
	struct shfs_directory *dir;
	char buf[SHFS_LINE_MAX];
	struct shfs_dir_list_node *file;
	int res;
	
	if (shfs_get_name(dentry->d_parent, buf) < 0)
		return -ENAMETOOLONG;
	res = shfs_dcache_get(info, buf, &dir);

	if (res < 0) {
		VERBOSE(" shfs_cache_get failed!\n");
		return res;
	}

	for (file = dir->head; file != NULL; file = file->next)
		if (strcmp(dentry->d_name.name, file->entry.name) == 0)
			break;

	if (!file) {
		DEBUG(" file not found in parent dir cache!\n");
		return -ENOENT;
	}

	fattr->f_mode = file->entry.mode;
/*	if (S_ISDIR(file->entry.mode))
		fattr->f_mode |= info->mnt.dir_mode;
	else if (S_ISREG(file->entry.mode))
		fattr->f_mode |= info->mnt.file_mode;
*/
	
	fattr->f_ino = file->entry.ino;
	fattr->f_size = file->entry.size;
	fattr->f_rdev = file->entry.rdev;
	fattr->f_blksize = file->entry.blocksize;
	fattr->f_blocks = file->entry.blocks;
	fattr->f_nlink = file->entry.nlink;
	fattr->f_atime = file->entry.atime;

	/*Any way to get the real mtime, ctime?*/
	fattr->f_mtime = file->entry.atime;
	fattr->f_ctime = file->entry.atime;
	if (info->mnt.preserve_own) {
		fattr->f_uid = file->entry.uid;
		fattr->f_gid = file->entry.gid;
	} else {
		fattr->f_uid = info->mnt.uid;
		fattr->f_gid = info->mnt.gid;
	}
	return 0;
}

/* data should belaigned (offset % count == 0) */
int
shfs_proc_read(struct dentry *dentry, unsigned offset, unsigned count, char *buffer){
	struct inode *inode = dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	struct shfs_sb_info *info = (struct shfs_sb_info*)sb->u.generic_sbp;
	char file[SHFS_PATH_MAX + 6];
	char line[SHFS_LINE_MAX];
	int res;

	if (shfs_get_name(dentry, file) < 0)
		return -ENAMETOOLONG;
	DEBUG(" %s, off %u, %u bytes\n", file, (unsigned)offset, (unsigned)count);
	if (count && offset % count)
		VERBOSE("Reading unaligned data, could hang!\n");
	if (strlen(info->mnt.root) + strlen(file) + 1 > SHFS_PATH_MAX) {
		VERBOSE("Path too long!\n");
		return -1;
	}

	if ((res = do_read(info, file, offset, count)) < 0) {
		VERBOSE("Can't send read command!\n");
		return res;
	}
	if ((res = shfs_pipe_readln(info, line, SHFS_LINE_MAX)) < 0) {
		VERBOSE("Can't see read response!\n");
		info->mnt.garbage = 1;
		info->mnt.garbage_read = count;
		return res;
	}

	switch (reply(line)) {
	case REP_PRELIM:
		break;
	case REP_EPERM:
		return -EPERM;
	case REP_ENOENT:
		return -ENOENT;
	default:
		info->mnt.garbage = 1;
		return -EIO;
	}
	if ((res = pipe_read(info, buffer, count)) < 0) {
		VERBOSE("Read: no data!\n");
		return res;
	}
	if ((res = shfs_pipe_readln(info, line, SHFS_LINE_MAX)) < 0) {
		VERBOSE("Can't see read response!\n");
		return res;
	}
	
	switch (reply(line)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	case REP_ENOENT:
		return -ENOENT;
	default:
		info->mnt.garbage = 1;
		return -EIO;
	}

	DEBUG(" received %u bytes\n", count);
	return count;
}

int
shfs_proc_write(struct dentry *dentry, unsigned offset, unsigned count, char *buffer){
	struct inode *inode = dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	struct shfs_sb_info *info = (struct shfs_sb_info*)sb->u.generic_sbp;
	char file[SHFS_PATH_MAX];
	char line[SHFS_LINE_MAX];
	int res;
	
	if (strlen(file) + strlen(info->mnt.root) + 1 > SHFS_PATH_MAX) {
		VERBOSE(" Path too long!\n");
		return -1;
	}

//	if (!count)
//		return 0;
	
	if (shfs_get_name(dentry, file) < 0)
		return -ENAMETOOLONG;
	DEBUG(" %s, off %u, %u bytes\n", file, offset, count);
		
	if ((res = do_write(info, file, offset, count)) < 0) {
		VERBOSE("Couldn't send write command!\n");
		goto out;
	}
	if ((res = shfs_pipe_readln(info, line, SHFS_LINE_MAX)) < 0) {
		VERBOSE("Cannot read response!\n");
		info->mnt.garbage = 1;
		info->mnt.garbage_write = count;
		goto out;
	}
	switch (reply(line)) {
	case REP_PRELIM:
		break;
	case REP_EPERM:
		res = -EPERM;
		goto out;
	case REP_ENOENT:
		res = -ENOENT;
		goto out;
	default:
		info->mnt.garbage = 1;
		res = -EIO;
		goto out;
	}

	if ((res = pipe_write(info, buffer, count)) < 0) {
		VERBOSE("Couldn't write file!\n");
		goto out;
	}
	if ((res = shfs_pipe_readln(info, line, SHFS_LINE_MAX)) < 0) {
		VERBOSE("Cannot read response!\n");
		goto out;
	}

	switch (reply(line)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		res = -EPERM;
		goto out;
	case REP_ENOENT:
		res = -ENOENT;
		goto out;
	case REP_ENOSPC:
		res = -ENOSPC;
		info->mnt.garbage = 1;
		info->mnt.garbage_write = count;
		goto out;
	default:
		info->mnt.garbage = 1;
		res = -EIO;
		goto out;
	}

	DEBUG(" sent %u bytes\n", count);
	/* shfs_dcache_invalidate is called in file.c */
	return count;
out:
	shfs_dcache_invalidate(dentry->d_parent);
	return res;
}

int
shfs_proc_mkdir(struct shfs_sb_info *info, char *dir){
	int res;
	char buf[SHFS_LINE_MAX];
	
	if (strlen(info->mnt.root) + strlen(dir) + 1 > SHFS_PATH_MAX) {
		VERBOSE("Path too long!\n");
		return -1;
	}

	DEBUG("Creating directory: %s%s\n", info->mnt.root, dir);
	
	if (!shfs_lock(info))
		return -EINTR;
	if ((res = do_mkdir(info, dir)) < 0) {
		VERBOSE("Could'n create directory!\n");
		shfs_unlock(info);
		return res;
	}

	shfs_pipe_readln(info, buf, SHFS_LINE_MAX);
	shfs_unlock(info);
	
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	default:
		return -EIO;
	}
	return 0;
}

int
shfs_proc_rmdir(struct shfs_sb_info *info, char *dir){
	char buf[SHFS_LINE_MAX];
	int res;
	
	if (strlen(info->mnt.root) + strlen(dir) + 1 > SHFS_PATH_MAX) {
		VERBOSE("Path too long!\n");
	        return -1;
	}

	DEBUG("Removing directory: %s%s\n", info->mnt.root, buf);
	
	if (!shfs_lock(info))
		return -EINTR;
	if ((res = do_rmdir(info, dir)) < 0) {
		VERBOSE("Couldn't remove directory!\n");
		shfs_unlock(info);
		return res;
	}

	shfs_pipe_readln(info, buf, SHFS_LINE_MAX);
	shfs_unlock(info);
	
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	case REP_ENOENT:
		return -ENOENT;
	default:
		return -EIO;
	}
	return 0;
}

int
shfs_proc_rename(struct shfs_sb_info *info, char *old, char *new){
	char buf[SHFS_LINE_MAX];
	int res;
	
	if ((strlen(info->mnt.root) + strlen(new) + 1 > SHFS_PATH_MAX) 
	   || (strlen(info->mnt.root) + strlen(old) + 1 > SHFS_PATH_MAX)) {
		VERBOSE("Path too long!\n");
	        return -1;
	}

	DEBUG("Rename %s%s -> %s%s\n", info->mnt.root, old, info->mnt.root, new);
	
	if (!shfs_lock(info))
		return -EINTR;
	if ((res = do_mv(info, old, new)) < 0) {
		VERBOSE("Couldn't rename!\n");
		shfs_unlock(info);
		return res;
	}

	shfs_pipe_readln(info, buf, SHFS_LINE_MAX);
	shfs_unlock(info);
	
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	case REP_ENOENT:
		return -ENOENT;
	default:
		return -EIO;
	}
	return 0;
}

int
shfs_proc_unlink(struct shfs_sb_info *info, char *file){
	char buf[SHFS_LINE_MAX];
	int res;
	
	if (strlen(info->mnt.root) + strlen(file) + 1 > SHFS_PATH_MAX) {
		VERBOSE("Path too long!\n");
	        return -1;
	}

	DEBUG("Removing file: %s%s\n", info->mnt.root, file);
	
	if (!shfs_lock(info))
		return -EINTR;
	if ((res = do_rm(info, file)) < 0) {
		VERBOSE("Couldn't remove file!\n");
		shfs_unlock(info);
		return res;
	}

	shfs_pipe_readln(info, buf, SHFS_LINE_MAX);
	shfs_unlock(info);
	
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	case REP_ENOENT:
		return -ENOENT;
	default:
		return -EIO;
	}
	return 0;
}

int
shfs_proc_create(struct shfs_sb_info *info, char *file, int mode)
{
	char buf[SHFS_LINE_MAX];
	int res;
	
	if (strlen(info->mnt.root) + strlen(file) + 1 > SHFS_PATH_MAX) {
		VERBOSE("Path too long!\n");
	        return -1;
	}
	
	DEBUG("Creating file: %s%s\n", info->mnt.root, file);
	
	if (!shfs_lock(info))
		return -EINTR;
	if ((res = do_creat(info, file, mode & S_IALLUGO)) < 0) {
		VERBOSE("Couldn't remove file!\n");
		shfs_unlock(info);
		return res;
	}

	shfs_pipe_readln(info, buf, SHFS_LINE_MAX);
	shfs_unlock(info);
	
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	default:
		return -EIO;
	}
	return 0;
}

int
shfs_proc_link(struct shfs_sb_info *info, char *old, char *new)
{
	char buf[SHFS_LINE_MAX];
	int res;
	
	if ((strlen(info->mnt.root) + strlen(new) + 1 > SHFS_PATH_MAX)
	   ||(strlen(info->mnt.root) + strlen(old) + 1 > SHFS_PATH_MAX)) {
		VERBOSE("Path too long!\n");
	        return -1;
	}

	DEBUG("Link %s%s -> %s%s\n", info->mnt.root, old, info->mnt.root, new);
	
	if (!shfs_lock(info))
		return -EINTR;
	if ((res = do_ln(info, old, new)) < 0) {
		VERBOSE("Couldn't make link!\n");
		shfs_unlock(info);
		return res;
	}

	shfs_pipe_readln(info, buf, SHFS_LINE_MAX);
	shfs_unlock(info);
	
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	default:
		return -EIO;
	}
	return 0;
}

int
shfs_proc_symlink(struct shfs_sb_info *info, const char *old, const char *new)
{
	char buf[SHFS_LINE_MAX];
	int res;
	
	if ((strlen(info->mnt.root) + strlen(new) + 1 > SHFS_PATH_MAX)
	   ||(strlen(info->mnt.root) + strlen(old) + 1 > SHFS_PATH_MAX)) {
		VERBOSE("Path too long!\n");
	        return -1;
	}

	DEBUG("Symlink %s%s -> %s\n", info->mnt.root, old, new);
	
	if (!shfs_lock(info))
		return -EINTR;
	if ((res = do_sln(info, old, new)) < 0) {
		VERBOSE("Couldn't make symlink!\n");
		shfs_unlock(info);
		return res;
	}

	shfs_pipe_readln(info, buf, SHFS_LINE_MAX);
	shfs_unlock(info);
	
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	default:
		return -EIO;
	}
	return 0;
}

int
shfs_proc_chmod(struct shfs_sb_info *info, char *file, umode_t mode)
{
	char buf[SHFS_LINE_MAX];
	int res;
	
	if (strlen(info->mnt.root) + strlen(file) + 1 > SHFS_PATH_MAX) {
		VERBOSE("Path too long!\n");
	        return -1;
	}

	DEBUG("Chmod %o %s%s\n", mode, info->mnt.root, file);
	
	if (!shfs_lock(info))
		return -EINTR;
	if ((res = do_chmod(info, file, mode & S_IALLUGO)) < 0) {
		VERBOSE("Couldn't do chmod!\n");
		shfs_unlock(info);
		return res;
	}

	shfs_pipe_readln(info, buf, SHFS_LINE_MAX);
	shfs_unlock(info);
	
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	case REP_ENOENT:
		return -ENOENT;
	default:
		return -EIO;
	}
	return 0;
}

int
shfs_proc_chown(struct shfs_sb_info *info, char *file, uid_t user)
{
	char buf[SHFS_LINE_MAX];
	int res;
	
	if (strlen(info->mnt.root) + strlen(file) + 1 > SHFS_PATH_MAX) {
		VERBOSE("Path too long!\n");
	        return -1;
	}

	DEBUG("Chown %u %s%s\n", user, info->mnt.root, file);
	
	if (!shfs_lock(info))
		return -EINTR;
	if ((res = do_chown(info, file, user)) < 0) {
		VERBOSE("Couldn't do chown!\n");
		shfs_unlock(info);
		return res;
	}

	shfs_pipe_readln(info, buf, SHFS_LINE_MAX);
	shfs_unlock(info);
	
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	case REP_ENOENT:
		return -ENOENT;
	default:
		return -EIO;
	}
	return 0;
}

int
shfs_proc_chgrp(struct shfs_sb_info *info, char *file, gid_t group)
{
	char buf[SHFS_LINE_MAX];
	int res;
	
	if (strlen(info->mnt.root) + strlen(file) + 1 > SHFS_PATH_MAX) {
		VERBOSE("Path too long!\n");
	        return -1;
	}

	DEBUG("Chown %u %s%s\n", group, info->mnt.root, file);
	
	if (!shfs_lock(info))
		return -EINTR;
	if ((res = do_chgrp(info, file, group)) < 0) {
		VERBOSE("Couldn't do chgrp!\n");
		shfs_unlock(info);
		return res;
	}

	shfs_pipe_readln(info, buf, SHFS_LINE_MAX);
	shfs_unlock(info);
	
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	case REP_ENOENT:
		return -ENOENT;
	default:
		return -EIO;
	}
	return 0;
}

int
shfs_proc_trunc(struct shfs_sb_info *info, char *file, loff_t size)
{
	char buf[SHFS_LINE_MAX];
	int res;
	
	if (strlen(info->mnt.root) + strlen(file) + 1 > SHFS_PATH_MAX) {
		VERBOSE("Path too long!\n");
	        return -1;
	}

	DEBUG("Truncate %s%s %u\n", info->mnt.root, file, (unsigned)size);
	
	if (!shfs_lock(info))
		return -EINTR;
	if ((res = do_trunc(info, file, size)) < 0) {
		VERBOSE("Couldn't do truncate!\n");
		shfs_unlock(info);
		return res;
	}

	shfs_pipe_readln(info, buf, SHFS_LINE_MAX);
	shfs_unlock(info);
	
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	case REP_ENOENT:
		return -ENOENT;
	default:
		return -EIO;
	}
	return 0;
}

int 
shfs_get_system(struct shfs_sb_info *info)
{	
	char buf[SHFS_LINE_MAX];
	char *s, *r, *system, *release;
	int i;
	int res;
	
	DEBUG("Uname\n");
	
	if (!shfs_lock(info))
		return -EINTR;
	if ((res = do_uname(info)) < 0) {
		VERBOSE("Couldn't get uname!\n");
		goto error;
	}

	if ((res = shfs_pipe_readln(info, buf, SHFS_LINE_MAX)) < 0) {
		VERBOSE("Error reading response!\n");
		goto error;
	}

	DEBUG("system: %s\n", buf);
	if ((r = strchr(buf, ' ')))
		*(r++) = '\0';
	s = buf;
	i = 0;
	system = shfs_system[i].system;
	release = shfs_system[i].release;
	while (system || release) {
		if (!info->mnt.generic_host && !strncmp(s, system, strlen(system))) {
			if (!release || (!strncmp(r, release, strlen(release)))) 
				break;
		}
		i++;
		system = shfs_system[i].system;
		release = shfs_system[i].release;
	}

	if (system)
		printk(KERN_INFO "shfs: remote peer: %s (%s)\n", system, release ? release : "generic");
	else
		printk(KERN_INFO "shfs: remote peer: generic\n");

	shfs_pipe_readln(info, buf, SHFS_LINE_MAX);
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		res = -EPERM;
		goto error;
	default:
		res = -EIO;
		goto error;
	}

	if ((res = do_init(info, i)) < 0) {
		VERBOSE("Cannot send init!\n");
		goto error;
	}

	if (shfs_pipe_readln(info, buf, SHFS_LINE_MAX) < 0) {
		VERBOSE("Error reading response!\n");
		goto error;
	}
	shfs_unlock(info);
	switch (reply(buf)) {
	case REP_COMPLETE:
		break;
	case REP_EPERM:
		return -EPERM;
	default:
		return -EIO;
	}
	return 0;
error:	
	shfs_unlock(info);
	return res;
}

int
shfs_notify_change(struct dentry *dentry, struct iattr *attr)
{
	struct inode *inode = dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	struct shfs_sb_info *info = (struct shfs_sb_info*)sb->u.generic_sbp;
	char file[SHFS_PATH_MAX];
	int res = 0;

	if (shfs_get_name(dentry, file) < 0)
		return -ENAMETOOLONG;

	DEBUG(" \n");

	if (attr->ia_valid & ATTR_MODE) {
		if ((res = shfs_proc_chmod(info, file, attr->ia_mode)) < 0)
			goto error;
	}
	if (attr->ia_valid & ATTR_UID) {
		if ((res = shfs_proc_chown(info, file, attr->ia_uid)) < 0)
			goto error;
	}
	if (attr->ia_valid & ATTR_GID) {
		if ((res = shfs_proc_chgrp(info, file, attr->ia_gid)) < 0)
			goto error;
	}
	if (attr->ia_valid & ATTR_SIZE) {
		if ((res = shfs_proc_trunc(info, file, attr->ia_size)) < 0)
			goto error;
		if (!shfs_lock(info))
			return -EINTR;
                inode->i_size = 0;
                mark_inode_dirty(inode);
		shfs_dcache_update_fattr(dentry, attr);
		shfs_unlock(info);
		return res;
	}
error:
	if (!shfs_lock(info))
		return -EINTR;
	shfs_dcache_update_fattr(dentry, attr);
	shfs_unlock(info);
	return res;
}


